import pandas as pd
from pathlib import Path
from datetime import datetime


class Reader:
    def __init__(self):
        self.required_headers = {
            'patient_id': 'patient_id',
            'study_date': 'study_date',
            'measurement': 'measurement',
            'alias': 'lesion_label_alias',
            'category': 'lesion_category'
        }
        self._categories = {
            'target': 'target',
            'non-target': 'non-target'
        }
        self._df = pd.DataFrame()

    @property
    def df(self):
        return self._df

    def _is_valid_date_format(self, date: str) -> bool:
        try:
            datetime.strptime(str(date), "%Y%m%d")
            return True
        except ValueError:
            return False

    def _check_format(self, df: pd.DataFrame) -> pd.DataFrame:
        # Check headers
        missing = set(self.required_headers.values()) - set(df.columns)
        if missing:
            raise ValueError(f"Missing required columns: {missing}")
        # Check format of study dates
        invalid_dates = df[self.required_headers['study_date']].dropna().astype(str).apply(lambda x: not self._is_valid_date_format(x))
        if invalid_dates.any():
            bad_rows = df.loc[invalid_dates, self.required_headers['study_date']].head(5).tolist()
            raise ValueError(f"Invalid date format. Expected 'YYYYMMDD'. Given: {bad_rows}")
        # Check categories
        invalid_categories = df[~df[self.required_headers['category']].isin(self._categories.values())]
        if not invalid_categories.empty:
            bad_rows = invalid_categories[self.required_headers['category']].head(5).tolist()
            raise ValueError(f"Invalid categories. Expected: {','.join(self._categories.values())}. Given: {bad_rows}")
        return df

    def read_measurements(self, path_to_measurements: str):
        """Update 'df' attribute by reading the CSV file in 'path_to_measurements'."""
        path_to_data = Path(path_to_measurements)
        if path_to_data.suffix.lower() != '.csv':
            raise ValueError(f"Expected a .csv file, got '{path_to_data.suffix}'")
        df = pd.read_csv(path_to_data)
        df = self._check_format(df)
        self._df = df



import pandas as pd

from pyrecist.data import Reader


class Evaluator:
    def __init__(self, reader: Reader):
        self.reader = self._check_reader(reader)
        self._response_classes = {
            'SD': 'SD',
            'PR': 'PR',
            'PD': 'PD',
            'CR': 'CR'
        }
        self._tumoral_burden_header = 'tumoral_burden'

    def _check_reader(self, reader: Reader) -> Reader:
        if reader.df.empty:
            raise ValueError(f"Expected a Reader with a dataframe loaded. Given: {reader.df}. Please run 'reader.read_measurements' before.")
        return reader

    def evaluate_recist(self) -> pd.DataFrame:
        patient_id_header = self.reader.required_headers['patient_id']
        study_date_header = self.reader.required_headers['study_date']
        measurement_header = self.reader.required_headers['measurement']
        classifications = []
        # Compute total tumor burden
        tumoral_burden_df = self.reader.df.groupby([patient_id_header, study_date_header])[measurement_header].sum().reset_index(name=self._tumoral_burden_header)
        # Remove patients with only baseline
        study_dates = tumoral_burden_df.groupby([patient_id_header])[study_date_header].size().reset_index(name="date_counts")
        cleaned_study_dates = study_dates[study_dates["date_counts"] > 1]
        tumoral_burden_df = tumoral_burden_df[tumoral_burden_df[patient_id_header].isin(cleaned_study_dates[patient_id_header].unique().tolist())]
        tumoral_burden_df.sort_values(by=[patient_id_header, study_date_header], inplace=True)
        # Loop over patients
        for patient_id in tumoral_burden_df[patient_id_header].unique().tolist():
            # Compute differences between subsequen studies
            studies_subset_df = tumoral_burden_df[tumoral_burden_df[patient_id_header] == patient_id].sort_values(by=study_date_header).reset_index()
            difference_df = studies_subset_df.diff()
            difference_df.dropna(inplace=True)
            difference_df[patient_id_header] = studies_subset_df[patient_id_header].iloc[1:]
            difference_df[study_date_header] = studies_subset_df[study_date_header].iloc[1:]
            # Set initial values
            baseline_date = studies_subset_df[study_date_header].iloc[0]
            nadir_date = studies_subset_df[study_date_header].iloc[0]
            baseline_burden = studies_subset_df[self._tumoral_burden_header].iloc[0]
            min_burden = studies_subset_df[self._tumoral_burden_header].iloc[0]
            reach_PD = False
            reach_PR = False
            # Loop over follow-up studies
            for idx, (_, row) in enumerate(difference_df.iterrows()):
                current_burden = studies_subset_df[self._tumoral_burden_header].iloc[idx+1]
                percentual_variation_from_nadir = 100 * (current_burden - min_burden) / min_burden
                percentual_variation_from_baseline = 100 * (current_burden - baseline_burden) / baseline_burden
                if not reach_PD:
                    if not reach_PR:
                        if percentual_variation_from_nadir >= 20.0 and (current_burden - min_burden >= 5):
                            classification = self._response_classes["PD"]
                            reach_PD = True
                        elif percentual_variation_from_baseline <= -30.0:
                            classification = self._response_classes["PR"]
                            reach_PR = True
                        else:
                            classification = self._response_classes["SD"]
                    else:
                        if percentual_variation_from_nadir >= 20.0 and (current_burden - min_burden >= 5):
                            classification = self._response_classes["PD"]
                            reach_PD = True
                        else:
                            classification = self._response_classes["PR"] # once 'PR' is reached, stays the same until 'PD' or 'CR'
                else:
                    classification = None
                # Add classification
                classifications.append({
                    "patient_id": patient_id,
                    "current_date": str(int(row[study_date_header])), # Why str(int())
                    "baseline_date": baseline_date,
                    "nadir_date": nadir_date,
                    "current_burden": current_burden,
                    "baseline_burden": baseline_burden,
                    "min_burden": min_burden,
                    "burden_difference_from_baseline": current_burden - baseline_burden,
                    "burden_percentual_variation_from_baseline": percentual_variation_from_baseline,
                    "burden_difference_from_nadir": current_burden - min_burden,
                    "burden_percentual_variation_from_nadir": percentual_variation_from_nadir,
                    "classification": classification
                })
                # Update nadir
                if current_burden < min_burden:
                    min_burden = current_burden
                    nadir_date = row[study_date_header].item()
        classifications_df = pd.DataFrame(classifications)
        return classifications_df



import sys
import argparse
import traceback
from pathlib import Path

from pyrecist.data import Reader
from pyrecist.assessment import Evaluator


def main():
    parser = argparse.ArgumentParser(
        description="Evaluate RECIST criteria from lesion measurement CSV.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        'path_to_csv',
        type=str,
        help="""Path to the .csv file containing the lesions measurements.
        Please, see the README.md to check the column headers and expected
        format."""
    )
    parser.add_argument(
        '-o',
        '--output',
        type=str,
        default=str(Path.cwd()),
        dest='path_to_output',
        help="Path to the output directory to save results."
    )
    parser.add_argument(
        '--patient_id_header',
        type=str,
        default='patient_id',
        help="Column header to identify patient ID."
    )
    parser.add_argument(
        '--study_date_header',
        type=str,
        default='study_date',
        help="Column header to identify study dates."
    )
    parser.add_argument(
        '--measurement_header',
        type=str,
        default='measurement',
        help="Column header to identify measurements."
    )
    parser.add_argument(
        '--alias_header',
        type=str,
        default='lesion_label_alias',
        help="""Column header to identify the alias (like an ID) for each
        lesion within a patient."""
    )
    parser.add_argument(
        '--category_header',
        type=str,
        default='lesion_category',
        help="""Column header to identify lesion categories (target, non-target)."""
    )
    args = parser.parse_args()
    try:
        reader = Reader()
        reader.required_headers.update({
            'patient_id': args.patient_id_header,
            'study_date': args.study_date_header,
            'measurement': args.measurement_header,
            'alias': args.alias_header,
            'category': args.category_header
        })
        reader.read_measurements(args.path_to_csv)
        evaluator = Evaluator(reader)
        results_df = evaluator.evaluate_recist()
        path_to_output_file = Path(args.path_to_output) / 'recist_evaluation.csv'
        results_df.to_csv(
            path_to_output_file,
            index=False
        )
        print(f"Results succesfully saved to {path_to_output_file}.")
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        traceback.print_exc()
        sys.exit(1)


if __name__ == '__main__':
    main()


import pytest
import pandas as pd
from pathlib import Path

from pyrecist.data import Reader


def test_reader_valid_headers(tmp_path):
    csv = tmp_path / "test.csv"
    csv_content = (
        "patient_id,study_date,measurement,lesion_label_alias,lesion_category\n"
        "1,20220929,23.9,A,target\n"
    )
    csv.write_text(csv_content)
    csv_df = pd.read_csv(csv)
    reader = Reader()
    reader.required_headers.update(
        {
            'patient_id': 'patient_id',
            'study_date': 'study_date',
            'measurement': 'measurement',
            'alias': 'lesion_label_alias',
            'category': 'lesion_category'
        }
    )
    assert isinstance(reader._check_format(csv_df), pd.DataFrame)


def test_reader_invalid_headers(tmp_path):
    csv = tmp_path / "test.csv"
    csv_content = (
        "patient_id,study_date,measurement,alias,lesion_category\n"
        "1,20220929,23.9,A,target\n"
    )
    csv.write_text(csv_content)
    csv_df = pd.read_csv(csv)
    reader = Reader()
    reader.required_headers.update(
        {
            'patient_id': 'patient_id',
            'study_date': 'study_date',
            'measurement': 'measurement',
            'alias': 'lesion_label_alias',
            'category': 'lesion_category'
        }
    )
    with pytest.raises(ValueError):
        reader._check_format(csv_df)


def test_reader_invalid_categories(tmp_path):
    csv = tmp_path / "test.csv"
    csv_content = (
        "patient_id,study_date,measurement,lesion_label_alias,lesion_category\n"
        "1,20220929,23.9,A,measurable\n"
    )
    csv.write_text(csv_content)
    csv_df = pd.read_csv(csv)
    reader = Reader()
    reader.required_headers.update(
        {
            'patient_id': 'patient_id',
            'study_date': 'study_date',
            'measurement': 'measurement',
            'alias': 'lesion_label_alias',
            'category': 'lesion_category'
        }
    )
    with pytest.raises(ValueError):
        reader._check_format(csv_df)


def test_reader_synthetic_data():
    csv = Path(__file__).parent / "data" / "synthetic_measurements.csv"
    csv_df = pd.read_csv(csv)
    reader = Reader()
    reader.required_headers.update(
        {
            'patient_id': 'patient_id',
            'study_date': 'study_date',
            'measurement': 'measurement',
            'alias': 'lesion_label_alias',
            'category': 'lesion_category'
        }
    )
    reader.read_measurements(csv)
    assert isinstance(reader.df, pd.DataFrame)
    assert not reader.df.empty
    assert csv_df.shape == reader.df.shape
    assert (csv_df.columns == reader.df.columns).all()


import pandas as pd
from pathlib import Path

from pyrecist.data import Reader
from pyrecist.assessment import Evaluator


def test_evaluator_synthetic_measurements():
    # Read synthetic data
    path_to_measurements = Path(__file__).parent / "data" / "synthetic_measurements.csv"
    reader = Reader()
    reader.required_headers.update(
        {
            'patient_id': 'patient_id',
            'study_date': 'study_date',
            'measurement': 'measurement',
            'alias': 'lesion_label_alias',
            'category': 'lesion_category'
        }
    )
    reader.read_measurements(path_to_measurements)
    # Run RECIST evaluation
    evaluator = Evaluator(reader)
    output_evaluation_df = evaluator.evaluate_recist()
    ground_truth_evaluation_df = pd.read_csv(Path(__file__).parent / "data" / "synthetic_evaluation.csv")
    # Check results
    assert isinstance(output_evaluation_df, pd.DataFrame)
    assert output_evaluation_df.shape == ground_truth_evaluation_df.shape
    assert (output_evaluation_df.columns == ground_truth_evaluation_df.columns).all()
    # Compare columns
    date_columns = [
        'current_date',
        'baseline_date',
        'nadir_date'
    ]
    float_columns = [
        'current_burden',
        'baseline_burden',
        'min_burden',
        'burden_difference_from_baseline',
        'burden_percentual_variation_from_baseline',
        'burden_difference_from_nadir',
        'burden_percentual_variation_from_nadir'
    ]
    for column in date_columns:
        assert output_evaluation_df[column].astype(int).astype(str).equals(ground_truth_evaluation_df[column].astype(int).astype(str))
    for column in float_columns:
        assert output_evaluation_df[column].astype(float).round(3).equals(ground_truth_evaluation_df[column].astype(float).round(3))
    assert output_evaluation_df['classification'].equals(ground_truth_evaluation_df['classification'])
